import kotlinx.coroutines.*
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock
import java.lang.StringBuilder
import java.util.concurrent.atomic.AtomicInteger
import kotlin.concurrent.thread
import kotlin.streams.toList
import kotlin.system.measureTimeMillis
import kotlin.time.ExperimentalTime

@ExperimentalTime
fun main() {
    helloCoroutines()
    // testAsync()
    // testParallelList()
    // testRunBlocking()
    //
    // testCancel1()
    // testCancel2()
    // testCancel3()
    // testCancel4()
    // testAsyncCancel()
    // testTimeout()
    //
    // testException1()
    // testException2()
    //
    // sharedData1()   // um den Thread-Pool einzurichten
    // sharedData1()   // erst ab jetzt sind die Benchmarktests aussagekräftig
    // sharedData2()
    // sharedData3()
    // sharedData4()
    // threadsVersusCoroutines()
    //
    // testSuspending()
}

fun helloCoroutines() {
    // launch startet eine Coroutine im Hintergrund
    GlobalScope.launch {
        delay(1000L)
        println("Coroutines!")
    }
    println("Thread-Anzahl: ${Thread.activeCount()}")
    // der Code hier wird sofort ausgeführt
    print("Hello ")
    Thread.sleep(2000L)  // sonst endet das Programm, bevor
    //                      der mit launch gestartete Code
    //                      fertig ist
}

fun testAsync() {
    val longtxt = "lorem ipsum"
    val result = GlobalScope.async(Dispatchers.IO) {
        longtxt.toSet().count()
    }
    println(result.isActive)
    Thread.sleep(200L)
    println(result.isActive)
    runBlocking {
        println(result.await())
    }

}

fun testParallelList() {
    val lst = List(100) { (0..100000).random() }
    var result = listOf<Int>()
    runBlocking {
        result = lst.map {
            GlobalScope.async(Dispatchers.Default) { it * 2 }
        }.awaitAll()
    }
    println(result)

    result = lst.parallelStream().map { it * 2}.toList()
    println(result)
}

fun testRunBlocking() {

    runBlocking {
        println("Im Block 1, Position 1")
        launch(Dispatchers.IO) {
            println("Ko1 startet, ${Thread.currentThread().name}")
            delay(2000L)
            println("Ko1 fertig")
        }
        launch(Dispatchers.Default) {
            println("Ko2 startet, ${Thread.currentThread().name}")
            delay(1000L)
            println("Ko2 fertig")
        }
        println("Im Block, Position 2")
        delay(500L)
        println("Im Block, Position 3")
    }
    println("Block 1 fertig")

    println("----")
    val myscope = CoroutineScope(Dispatchers.Default)
    runBlocking(myscope.coroutineContext) {
        println("Im Block 2, Position 1")
        launch(Dispatchers.IO) {
            println("Ko1 startet, ${Thread.currentThread().name}")
            delay(2000L)
            println("Ko1 fertig")
        }
        launch {  // automatisch in Dispatchers.Default
            println("Ko2 startet, ${Thread.currentThread().name}")
            delay(1000L)
            println("Ko2 fertig")
        }
        println("Im Block, Position 2")
        delay(500L)
        println("Im Block, Position 3")
    }
    println("Block 2 fertig")

    println("-----")
    runBlocking {
        repeat(10) {
            launch(Dispatchers.Default) {
                delay((0..500L).random())
                println("Koroutine $it")
            }
        }
    }
    println("Block 3 fertig")



}

// cancel bei delay
fun testCancel1() {
    val job = GlobalScope.launch {
        // die folgende Schleife im Hintergrund ausführen ...
        for(i in (1..10)) {
            val progress = "*".repeat(i)
            println("Fortschritt: $progress")
            delay((500..2000L).random())
        }
    }
    // ... und nach fünf Sekunden abbrechen
    Thread.sleep(5000)
    job.cancel()
    println("Ende")
    Thread.sleep(5000)
    // Ausgabe:
    // Fortschritt: *
    // Fortschritt: **
    // Fortschritt: ***
    // Fortschritt: ****
    // Ende
}

// cancel scheitert weil nicht kooperativ
fun testCancel2() {
    val job = GlobalScope.launch {
        // die folgende Schleife im Hintergrund ausführen ...
        for(i in (1..10)) {
            val progress = "*".repeat(i)
            println("Fortschritt: $progress")
            var sum = 0L
            for (j in (0..500_000_000))
                sum += j
        }
    }
    // ... und nach zwei Sekunden abbrechen
    Thread.sleep(2000)
    println("Cancel")
    job.cancel()
    println("Ende")
    Thread.sleep(5000)
    // Ausgabe:
    // Fortschritt: *
    // Fortschritt: **
    // Fortschritt: ***
    // Fortschritt: ****
    // Ende
    // Fortschritt: *****
    // Fortschritt: ******
    // Fortschritt: *******
    // ...
}

// kooperativ
fun testCancel3() {
    val job = GlobalScope.launch {
        // die folgende Schleife im Hintergrund ausführen ...
        for(i in (1..10)) {
            if (!isActive)
                break
            val progress = "*".repeat(i)
            println("Fortschritt: $progress")
            var sum = 0L
            for (j in (0..500_000_000))
                sum += j
        }
    }
    // ... und nach zwei Sekunden abbrechen
    Thread.sleep(2000)
    println("Cancel")
    job.cancel()
    println("Ende")
    Thread.sleep(5000)
    // Ausgabe:
    // Fortschritt: *
    // Fortschritt: **
    // Fortschritt: ***
    // Fortschritt: ****
    // Cancel
    // Ende
}

// mit try/finally
fun testCancel4() {
    val job = GlobalScope.launch {
        // die folgende Schleife im Hintergrund ausführen ...
        try {
            for (i in (1..10)) {
                val progress = "*".repeat(i)
                println("Fortschritt: $progress")
                delay(500L)
            }
        } catch (e: Exception) {
            println("Exception $e")
        }
        finally {
            // Aufräumarbeiten durchführen
            println("Finally")
        }
    }
    // ... und nach zwei Sekunden abbrechen
    Thread.sleep(2000)
    println("Cancel")
    job.cancel()
    println("Ende")
    Thread.sleep(5000)
    // Ausgabe:
    // Fortschritt: *
    // Fortschritt: **
    // Fortschritt: ***
    // Fortschritt: ****
    // Cancel
    // Ende
}

fun testAsyncCancel() {
    val result = GlobalScope.async(Dispatchers.Default) {
        var tmp = 0L
        // reagiert nicht auf cancel-Aufforderung
        for (i in (0..500_000_000L))
            tmp += i
        tmp  // Lambda-Ausdruck gibt result zurück
    }
    Thread.sleep(50L)
    println("Cancel")
    result.cancel()
    Thread.sleep(1000L)
    println(result.isCompleted)  // true
    println(result.isCancelled)  // true
    runBlocking {
        // löst Exception aus
        // println(result.await())
    }
    println("Programmende")
}

fun testTimeout() {
    runBlocking {
        try {
            withTimeout(1000L) {
                for (i in (1..10)) {
                    println("Fortschritt: $i")
                    delay(250L)
                }
            }
        } catch (e: Exception) {
            println("Exception $e")
        }
        println("weiter")

        // result hat den Datentyp Int?
        val result = withTimeoutOrNull(1000L) {
            delay((500..1500L).random())
            15
        }
        println(result)  // Ausgabe 15 oder null
    }
}

// Liste von Deferred-Elementen mit Exceptions auswerten
fun testException1() {
    println("Die folgenden Fehlermeldungen sind zu erwarten.")
    val results = mutableListOf<Deferred<Int>>()
    runBlocking {
        for (i in 1..100) {
            results += GlobalScope.async {
                val tmp = (0..110).random()
                if (tmp > 100)
                    throw ArithmeticException("Wert zu groß")
                tmp
            }
        }
        println("Auswertung 1:")
        try {
            val sum = results.awaitAll().sum()
            println("  Summe: $sum")
        } catch (e: Exception) {
            println("  Fehler $e")
        }
        println("Auswertung 2:")
        var sum = 0
        for (result in results) {
            try {
                sum += result.await()
            } catch (e: Exception) {
                println("  Fehler $e")
            }
        }
        println("  Summe aller gültigen Werte: $sum")

    }
    println("Weiter")
}

// Beispiel für CoroutineExceptionHandler
fun testException2() {
    println("Die folgenden Fehlermeldungen sind zu erwarten.")
    val myhandler = CoroutineExceptionHandler {
            context, exception ->
        println("Fehler vom Handler verarbeitet: $exception")
    }
    runBlocking {
        GlobalScope.launch(myhandler) {
            throw Exception("Ein Fehler im asynchronen Code.")
        }
    }
    println("Weiter")
    Thread.sleep(1000)  // sonst ist das Programm zuende,
                        // bevor der Fehler erscheint
}

// fehlerhaftes Ergebnis, weil sharedCounter nicht thread-safe ist
fun sharedData1() {
    val dur = measureTimeMillis {
        var sharedCounter = 0
        runBlocking {
            for (i in 1..250_000) {
                launch(Dispatchers.Default) {
                    delay((0..100L).random())
                    sharedCounter += 1
                }
            }
        }
        println(sharedCounter)
    }
    println("Zeit  sharedData1: $dur ms")
}

// thread-safe dank AtomicInteger
fun sharedData2() {
    val dur = measureTimeMillis {
        val sharedCounter = AtomicInteger()
        runBlocking {
            for (i in 1..250_000) {
                launch(Dispatchers.Default) {
                    delay((0..100L).random())
                    sharedCounter.incrementAndGet()
                }
            }
        }
        println(sharedCounter)
    }
    println("Zeit sharedData2: $dur ms")
}

// thread-safe dank Single-Thread-Kontext (deprecated)
@ObsoleteCoroutinesApi
fun sharedData3() {
    val dur = measureTimeMillis {
        var sharedCounter = 0
        val single = newSingleThreadContext("mycontext")
        runBlocking {
            for (i in 1..250_000) {
                launch(Dispatchers.Default) {
                    delay((0..100L).random())
                    withContext(single) {
                        sharedCounter += 1
                    }
                }
            }
        }
        println(sharedCounter)
    }
    println("Zeit  sharedData3: $dur ms")
}

// thread-safe dank Mutex
fun sharedData4() {
    val dur = measureTimeMillis {
        val myMutex = Mutex()
        var sharedCounter = 0
        runBlocking {
            for (i in 1..250_000) {
                launch(Dispatchers.Default) {
                    delay((0..100L).random())
                    myMutex.withLock {
                        sharedCounter += 1
                    }
                }
            }
        }
        println(sharedCounter)
    }
    println("Zeit  sharedData4: $dur ms")
}


@ExperimentalTime
fun threadsVersusCoroutines() {
    val number = 250000

    // zur Initialisierung des Thread-Pools, damit
    // die folgenden Aufrufe von measureTimeMillis
    // aussagekräftig sind
    runBlocking {
        for (i in 1..number) {
            launch(Dispatchers.Default) {
                delay((0..100L).random())
            }
        }
    }

    // Koroutinen ...
    var dur = measureTimeMillis {
        val sharedCounter = AtomicInteger()
        runBlocking {
            for (i in 1..number) {
                launch(Dispatchers.Default) {
                    delay((0..100L).random())
                    sharedCounter.incrementAndGet()
                }
            }
        }
        println("Ergebnis: $sharedCounter")
    }
    println("Koroutinen: $dur ms")


    // ... versus Threads
    dur = measureTimeMillis {
        val sharedCounter = AtomicInteger()
        val threadlist = mutableListOf<Thread>()
        for (i in 1..number) {
            threadlist += thread {
                // Thread.sleep((0..100L).random())
                sharedCounter.incrementAndGet()
            }
        }
        threadlist.map { it.join() }
        println("Ergebnis: $sharedCounter")

    }
    println("Threads: $dur ms")

}


fun testSuspending() {
    runBlocking {
        val job = launch(Dispatchers.Default) {
            doSomething()
        }
        val s = async(Dispatchers.Default) { buildLongString() }
        job.join()
        val result = s.await().takeLast(100)
        println("Die letzten Zeichen einer sehr langen Zeichenkette: $result")
    }
}

suspend fun doSomething() {
    withContext(Dispatchers.IO) {
        var sum = 0L
        for (i in (0..500_000_000L))
            sum += i
        println("Summe: $sum")
    }
}

fun buildLongString(): String {
    val sb = StringBuilder()
    for (i in (0..100_000))
        sb.append(("0123456789abcdef".random()))
    return sb.toString()
}